/*
 * Copyright (C) 2017 Miguel Castiblanco
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import 'dart:async';
import 'dart:io';
import 'dart:convert';

/// If one or more request interceptors are set, they will be called before the
/// request is made in order to allow for modifications of the [HttpClientRequest]
/// See [SimpleHttpClient.addRequestInterceptor]
typedef Future<HttpClientRequest> RequestInterceptor(HttpClientRequest request);

/// This type is used in order to allow the interception of the response,
/// usually you will want to use this for logging or analytics. See
/// [SimpleHttpClient.addResponseInterceptor]
typedef Future<Null> ResponseInterceptor(
    HttpClientRequest request, Response response);

// Defines a certificate by its bytes
class SimpleCertificate {
  final List<int> bytes;

  SimpleCertificate(this.bytes);
}

/// Wraps `dart:io`'s [HttpClient] to provide an easy to use, yet flexible and
/// feature rich implementation of an Http Client.
///
///     SimpleHttpClient client = new SimpleHttpClient();
///     Response response = await client.get("http://localhost:8080/service");
///     if (HttpStatus.OK == response.statusCode) {
///       // OK
///     }
///
/// Set an authenticator to have it called if a request receives a [:401:] response,
/// or set an interceptor if you want to modify all the requests made with this
/// client (e.g adding a header). See [SimpleHttpClient.setAuthenticator],
/// [SimpleHttpClient.addRequestInterceptor], and [RequestInterceptor]. For logging
/// or analytics purposes, you may want to set a [ResponseInterceptor], see
/// [SimpleHttpClient.addResponseInterceptor].
///
/// Set a timeout [Duration] when creating the instance
/// `new SimpleHttpClient(timeout: new Duration(seconds: 7))`.
///
/// You can also configure it to accept any TLS certificate, add a custom
/// self certificate, or accept only certs generated by trusted authorities (default
/// behavior)
///
/// Check the documentation on the constructor and the public methods for more
/// details
class SimpleHttpClient {
  /// Used for the `user-agent` header
  final String userAgent;

  /// Whether or not the client should follow redirects. Default value is [:true:]
  final bool followRedirects;

  /// Max amount of redirects supported. Default value is [:5:]
  final int maxRedirects;

  /// If an authenticator is set, how many times should the client retry calling
  /// the authenticator and the service before failing. Default value is [:1:]
  final int maxAuthRetries;

  /// Timeout for the requests. Default value is [:10:] seconds
  final Duration timeout;

  HttpClient _httpClient;
  List<RequestInterceptor> _requestInterceptors = new List();
  List<ResponseInterceptor> _responseInterceptors = new List();
  RequestInterceptor _authenticate;
  final SecurityContext _securityContext;

  /// Creates an instance of [SimpleHttpClient].
  ///
  /// [userAgent] Will be used when sending the requests. Defaults to library name
  ///
  /// [allowAnyCert] Whether or not the app should accept any certificate
  /// during secure requests. Consider setting [trustedCertificates] instead if you
  /// are connecting to your own self-signed certificates.
  ///
  /// [timeout] Sets a [Duration] as the timeout for a request to happen. Default
  /// is [:10:] seconds.
  ///
  /// [followRedirects] Whether or not the client should follow redirects, used
  /// together with [maxRedirects]. [followRedirects] is true by default, while
  /// [maxRedirects] is [:5:] by default.
  ///
  /// [trustedCertificates] allows you to add a certificate file to the security
  /// context, allowing you to set your self signed certificate as trusted. Please
  /// try to use [trustedCertificates] instead of [allowAnyCert] if possible.
  SimpleHttpClient(
      {this.userAgent: "simple_http_client",
      allowAnyCert: false,
      this.timeout: const Duration(seconds: 10),
      this.followRedirects: true,
      this.maxRedirects: 5,
      this.maxAuthRetries: 1,
      List<SimpleCertificate> trustedCertificates})
      : _securityContext =
            (trustedCertificates != null) ? new SecurityContext() : null {
    if (trustedCertificates != null) {
      trustedCertificates.forEach((cert) {
        _securityContext.setTrustedCertificatesBytes(new List.from(cert.bytes));
      });
    }

    _httpClient = new HttpClient(context: _securityContext);
    _httpClient.badCertificateCallback = (_, __, ___) => allowAnyCert;
  }

  _setHeaders(HttpClientRequest request) {
    request.headers.removeAll("user-agent");
    request.headers.add("user-agent", userAgent);
  }

  _interceptRequest(HttpClientRequest request) async {
    if (_requestInterceptors.isNotEmpty) {
      for (RequestInterceptor interceptor in _requestInterceptors) {
        await interceptor(request);
      }
    }
  }

  _interceptResponse(HttpClientRequest request, Response response) async {
    if (_responseInterceptors.isNotEmpty) {
      for (ResponseInterceptor interceptor in _responseInterceptors) {
        await interceptor(request, response);
      }
    }
  }

  _setContentType(HttpClientRequest request, {ContentType contentType}) {
    if (contentType == null) {
      contentType = ContentType.json;
    }

    request.headers.contentType = contentType;
  }

  Future<Response> _execute(Future<HttpClientRequest> handler(),
      {bool authenticate: false,
      int requestNumber: 1,
      String body,
      Map<String, String> headers}) async {
    try {
      HttpClientRequest request = await handler();
      _setHeaders(request);
      request.followRedirects = followRedirects;
      request.maxRedirects = maxRedirects;

      if (headers != null) {
        headers.forEach((key, value) {
          request.headers.add(key, value);
        });
      }

      if (authenticate) await _authenticate(request);
      await _interceptRequest(request);

      if (body != null) {
        request.write(body);
      }

      HttpClientResponse response =
          await request.close().timeout(timeout, onTimeout: () {
        throw new TimeOutException();
      });

      var simpleResponse = await _read(response);
      await _interceptResponse(request, simpleResponse);

      if (HttpStatus.unauthorized == response.statusCode &&
          _authenticate != null &&
          requestNumber <= maxAuthRetries) {
        return _execute(handler,
            authenticate: true,
            requestNumber: requestNumber + 1,
            body: body,
            headers: headers);
      } else if (HttpStatus.unauthorized == response.statusCode) {
        throw new UnauthorizedException();
      }

      return simpleResponse;
    } on SocketException catch (_) {
      throw new CantConnectException();
    } on HandshakeException catch (_) {
      throw new TLSException();
    } on CertificateException catch (_) {
      throw new TLSException();
    }
  }

  Future<HttpClientRequest> _get(String url) async {
    return await _httpClient.getUrl(Uri.parse(url));
  }

  Future<HttpClientRequest> _post(String url, {ContentType contentType}) async {
    HttpClientRequest request = await _httpClient.postUrl(Uri.parse(url));
    _setContentType(request, contentType: contentType);
    return request;
  }

  Future<HttpClientRequest> _put(String url, {ContentType contentType}) async {
    HttpClientRequest request = await _httpClient.putUrl(Uri.parse(url));
    _setContentType(request, contentType: contentType);
    return request;
  }

  Future<HttpClientRequest> _delete(String url) async {
    return await _httpClient.deleteUrl(Uri.parse(url));
  }

  Future<Response> _read(HttpClientResponse response) async {
    StringBuffer body = new StringBuffer();
    Completer<String> completer = new Completer();

    response.transform(utf8.decoder).listen((data) {
      body.write(data);
    }, onDone: () {
      completer.complete(body.toString());
    });

    Map<String, String> headers = new Map();

    response.headers.forEach((key, values) {
      headers[key] = values.join(", ");
    });

    return new Response()
      ..statusCode = response.statusCode
      ..headers = headers
      ..body = await completer.future;
  }

  /// Set the authenticator for the client. If an authenticator is set,
  /// the library will call it when a response has the HTTP Status Code [:401:].
  /// The authenticator will be called with a request identical to the original
  /// one, and the authenticator has to modify that request in whichever way
  /// needed. After the authenticator is called, the request will be sent and
  /// the execution will continue normally. The authenticator will be called
  /// until [maxAuthRetries] is reached. After that, or if an authenticator was
  /// not set, an [UnauthorizedException] will be thrown.
  ///
  /// The authenticator will be called before any interceptor.
  void setAuthenticator(RequestInterceptor authenticator) {
    this._authenticate = authenticator;
  }

  /// Add a [RequestInterceptor]. Each interceptor will be called sequentially in the same
  /// order it was added, before executing the request, and after the authenticator
  /// if authentication was needed.
  void addRequestInterceptor(RequestInterceptor interceptor) {
    _requestInterceptors.add(interceptor);
  }

  /// Remove a request interceptor from the list
  void removeRequestInterceptor(RequestInterceptor interceptor) {
    _requestInterceptors.remove(interceptor);
  }

  /// ResponseInterceptor are called as soon as a response is obtained. Usually
  /// this type of interceptor is used for analytics or logging
  void addResponseInterceptor(ResponseInterceptor interceptor) {
    _responseInterceptors.add(interceptor);
  }

  /// Remove a response interceptor from the list
  void removeResponseInterceptor(ResponseInterceptor interceptor) {
    _requestInterceptors.remove(interceptor);
  }

  /// Execute a POST call.
  ///
  /// [url] of the resource
  ///
  /// [body] of the request
  ///
  /// [contentType] to be used
  ///
  /// [headers] if any
  /// Will return a [Response]
  Future<Response> post(String url, String body,
      {ContentType contentType, Map<String, String> headers}) async {
    var response = await _execute(() => _post(url, contentType: contentType),
        body: body, headers: headers);
    return response;
  }

  /// Execute a PUT call.
  ///
  /// [url] of the resource
  ///
  /// [body] of the request
  ///
  /// [contentType] to be used
  ///
  /// [headers] if any
  /// Will return a [Response]
  Future<Response> put(String url, String body,
      {ContentType contentType, Map<String, String> headers}) async {
    var response = await _execute(() => _put(url, contentType: contentType),
        body: body, headers: headers);
    return response;
  }

  /// Execute a GET call.
  ///
  /// [url] of the resource
  ///
  /// [headers] if any
  ///
  /// Will return a [Response]
  Future<Response> get(String url, {Map<String, String> headers}) async {
    var response = await _execute(() => _get(url), headers: headers);
    return response;
  }

  /// Execute a GET call.
  ///
  /// [url] of the resource
  ///
  /// [headers] if any
  ///
  /// Will return a [Response]
  Future<Response> delete(String url, {Map<String, String> headers}) async {
    var response = await _execute(() => _delete(url), headers: headers);
    return response;
  }
}

/// Contains the information of the response
///
/// [statusCode] is the HTTP Status Code
///
/// [body] is an already-loaded String containing the contents of the response
///
/// [headers] is a [Map] with each header and its values, take into account that
/// if a header has more than one value, the values will be separated by comma,
/// for example `'x-sample-header': 'first, second'`
class Response {
  int statusCode;
  String body;
  Map<String, String> headers;
}

/// Thrown if the response has a [:401:] status code, and there is not
/// authenticator set (see [SimpleHttpClient.setAuthenticator]).
///
/// Will also be thrown if [maxAuthRetries] is reached.
class UnauthorizedException implements Exception {
  String toString() {
    return "Unauthorized";
  }
}

/// Thrown if the client can't connect to the specified url
class CantConnectException implements Exception {
  String toString() {
    return "Can't connect to the server";
  }
}

/// There is an error with the certificate received from the server. If you
/// are using a self-signed certificate use [trustedCertificates] when creating
/// the instance of the [SimpleHttpClient] to set add your certificate to the
/// security context.
///
/// You can also use [allowAnyCert] if you know that the certificate is
/// self-signed but can't be added ahead of time. Use [allowAnyCert] only
/// if you know what you are doing.
class TLSException implements Exception {
  String toString() {
    return "SSL error in the given port";
  }
}

/// The timeout for a request has been reached.
class TimeOutException implements Exception {
  String toString() {
    return "Request timed out";
  }
}
